use anyhow::Result;
use celery::{error::TaskError, task::TaskResult};
use chrono::prelude::*;
use diesel::prelude::*;

use crate::db;

fn do_cleanup(db_conn: &mut db::Connection) -> Result<()> {
    let timegrid_ids = db::schema::timegrids::table
        .select(db::schema::timegrids::id)
        .filter(db::schema::timegrids::active.eq(false))
        .load::<i32>(db_conn)?;
    let timegrid_day_ids = db::schema::timegrid_days::table
        .select(db::schema::timegrid_days::id)
        .filter(db::schema::timegrid_days::timegrid_id.eq_any(&timegrid_ids))
        .load::<i32>(db_conn)?;

    diesel::delete(
        db::schema::timegrid_time_unit::table
            .filter(db::schema::timegrid_time_unit::timegrid_day_id.eq_any(&timegrid_day_ids)),
    )
    .execute(db_conn)?;
    diesel::delete(
        db::schema::timegrid_days::table
            .filter(db::schema::timegrid_days::id.eq_any(timegrid_day_ids)),
    )
    .execute(db_conn)?;
    diesel::delete(
        db::schema::timegrids::table.filter(db::schema::timegrids::id.eq_any(timegrid_ids)),
    )
    .execute(db_conn)?;

    let subst_query_ids = db::schema::substitution_queries::table
        .select(db::schema::substitution_queries::id)
        .filter(
            db::schema::substitution_queries::queried_at
                .lt(Local::now().naive_local() - chrono::Duration::hours(1)),
        )
        .load::<i64>(db_conn)?;
    let subst_ids = db::schema::substitutions::table
        .select(db::schema::substitutions::id)
        .filter(db::schema::substitutions::substitution_query_id.eq_any(&subst_query_ids))
        .load::<i64>(db_conn)?;

    diesel::delete(
        db::schema::substitution_classes::table
            .filter(db::schema::substitution_classes::substitution_id.eq_any(&subst_ids)),
    )
    .execute(db_conn)?;
    diesel::delete(
        db::schema::substitution_teachers::table
            .filter(db::schema::substitution_teachers::substitution_id.eq_any(&subst_ids)),
    )
    .execute(db_conn)?;
    diesel::delete(
        db::schema::substitution_subjects::table
            .filter(db::schema::substitution_subjects::substitution_id.eq_any(&subst_ids)),
    )
    .execute(db_conn)?;
    diesel::delete(
        db::schema::substitution_rooms::table
            .filter(db::schema::substitution_rooms::substitution_id.eq_any(&subst_ids)),
    )
    .execute(db_conn)?;
    diesel::delete(
        db::schema::substitutions::table.filter(db::schema::substitutions::id.eq_any(subst_ids)),
    )
    .execute(db_conn)?;
    diesel::delete(
        db::schema::substitution_queries::table
            .filter(db::schema::substitution_queries::id.eq_any(subst_query_ids)),
    )
    .execute(db_conn)?;

    Ok(())
}

#[celery::task]
pub async fn cleanup() -> TaskResult<()> {
    let db_conn = &mut match db::POOL.get() {
        Ok(x) => x,
        Err(e) => return Err(TaskError::UnexpectedError(format!("{:?}", e))),
    };
    if let Err(e) = do_cleanup(db_conn) {
        return Err(TaskError::UnexpectedError(format!("{:?}", e)));
    }

    Ok(())
}
